import os
import numpy as np
import pandas as pd
from functools import reduce
import pickle
import inspect
import importlib.util
from .datapoint import load_df_data_from_folder

def retrieve_name(var):
    """
    Gets the name of var. Does it from the out most frame inner-wards.
    :param var: variable to get name from.
    :return: string
    """
    for fi in reversed(inspect.stack()):
        names = [var_name for var_name, var_val in fi.frame.f_locals.items() if var_val is var and var_name[0] != '_']
        # print(names)
        if len(names) > 0:
            return names[0]
        

def convert_variables_to_pystirng(string_variables = [], 
                                  iterative_variables = [], 
                                  fn_variables = [], 
                                  prefix = ['import pandas as pd', 
                                            'import numpy as np']):
    L = prefix
    
    for i in string_variables:
        line = f'{retrieve_name(i)} = "{i}"'
        L.append(line)
        
    for i in iterative_variables:
        line = f'{retrieve_name(i)} = {i}'
        L.append(line)
        
    for i in fn_variables:
        line = f'{i.fn_string}'
        L.append(line)
        

    D_str = "\n\nMetaDict = {" + ',\n'.join(
                    [f'"{retrieve_name(i)}": {retrieve_name(i)}'
                 for i in string_variables + iterative_variables + fn_variables]
                ) + "}"
     
    python_strings = '\n\n'.join(L) + D_str
    
    return python_strings


def load_module_variables(file_path):

    # Create a module specification
    spec = importlib.util.spec_from_file_location("module_name", file_path)

    # Create a module from the specification
    module = importlib.util.module_from_spec(spec)

    # Load the module
    spec.loader.exec_module(module)
    return module


def save_to_pickle(obj, filename):
    with open(filename, 'wb') as f:  # write in binary mode
        pickle.dump(obj, f)

def load_from_pickle(filename):
    with open(filename, 'rb') as f:  # read in binary mode
        return pickle.load(f)
    

def get_df_whole_from_settings(RecName_Chain, 
                               RecTableName2FldColumns_Dict, 
                               rec_folder, 
                               RecName_Chain_to_RecName = {}):
   # (1) get df_prefix
    L = []

    RecNameID_Chain = [f'{i}ID' for i in RecName_Chain]
    for idx, RecName in enumerate(RecName_Chain):
        RecName_Full = RecName_Chain_to_RecName.get(RecName, RecName)
        df = load_df_data_from_folder(os.path.join(rec_folder, RecName_Full)) # [prefix_ids + focal_ids]
        df = df[RecNameID_Chain[:idx+ 1]].astype(str).drop_duplicates()
        L.append(df)

    df_prefix = reduce(lambda left, right: pd.merge(left, right, how = 'left'), L)
    
    for RecID in RecNameID_Chain:
        s = 'M' + pd.Series(df_prefix.index).astype(str)
        df_prefix[RecID] = df_prefix[RecID].fillna(s)
        
    # (2) get df_data
    RecLevel = RecName_Chain[-1]
    RecLevelID = f'{RecLevel}ID'

    L = []
    for RecName, FldList in RecTableName2FldColumns_Dict.items():
        df = load_df_data_from_folder(os.path.join(rec_folder, RecName))
        if FldList == 'ALL': FldList = list(df.columns)
        full_cols = [i for i in RecNameID_Chain if i not in FldList] + FldList
        full_cols = [i for i in full_cols if i in df.columns]
        df = df[full_cols].reset_index(drop = True)

        for RecID in RecNameID_Chain:
            if RecID in df.columns: df[RecID] = df[RecID].astype(str)

        if RecLevelID not in df.columns:
            on_cols = [i for i in df_prefix.columns if i in df.columns]
            df = pd.merge(df_prefix, df, on = on_cols, how = 'left')

        df = pd.DataFrame([{RecLevelID: RecLevelIDValue, RecName: df_input} 
                           for RecLevelIDValue, df_input in df.groupby(RecLevelID)])
        L.append(df)
        
    # Merge the dataframes in the list using reduce
    df_data = reduce(lambda left, right: pd.merge(left, right, on=RecLevelID, how = 'left'), L)

    # 3. get df_whole
    df_whole = pd.merge(df_prefix, df_data, on = RecLevelID, how = 'left')

    for RecName in RecTableName2FldColumns_Dict:
        ind = df_whole[RecName].isna()
        columns = df_whole[-ind][RecName].iloc[0].columns
        df_whole[RecName] = df_whole[RecName].apply(lambda x: x if type(x) == type(pd.DataFrame()) 
                                                    else pd.DataFrame(columns = columns))
    
    return df_whole



def get_df_individual_from_settings(RecName_Chain, 
                                    RecTableName2FldColumns_Dict, 
                                    Pat, 
                                    RecName_Chain_to_RecName = {}):
   # (1) get df_prefix
    L = []

    RecNameID_Chain = [f'{i}ID' for i in RecName_Chain]
    for idx, RecName in enumerate(RecName_Chain):
        RecName_Full = RecName_Chain_to_RecName.get(RecName, RecName)
        # df = load_df_data_from_folder(os.path.join(rec_folder, RecName_Full)) # [prefix_ids + focal_ids]
        df = Pat.get_df_rec(RecName_Full)
        df = df[RecNameID_Chain[:idx+ 1]].astype(str).drop_duplicates()
        L.append(df)

    df_prefix = reduce(lambda left, right: pd.merge(left, right, how = 'left'), L)
    
    for RecID in RecNameID_Chain:
        s = 'M' + pd.Series(df_prefix.index).astype(str)
        df_prefix[RecID] = df_prefix[RecID].fillna(s)
        
    # (2) get df_data
    RecLevel = RecName_Chain[-1]
    RecLevelID = f'{RecLevel}ID'

    L = []
    for RecName, FldList in RecTableName2FldColumns_Dict.items():
        # df = load_df_data_from_folder(os.path.join(rec_folder, RecName))
        df = Pat.get_df_rec(RecName)
        if FldList == 'ALL': FldList = list(df.columns)
        full_cols = [i for i in RecNameID_Chain if i not in FldList] + FldList
        full_cols = [i for i in full_cols if i in df.columns]
        df = df[full_cols].reset_index(drop = True)

        for RecID in RecNameID_Chain:
            if RecID in df.columns: df[RecID] = df[RecID].astype(str)

        if RecLevelID not in df.columns:
            on_cols = [i for i in df_prefix.columns if i in df.columns]
            df = pd.merge(df_prefix, df, on = on_cols, how = 'left')

        df = pd.DataFrame([{RecLevelID: RecLevelIDValue, RecName: df_input} 
                           for RecLevelIDValue, df_input in df.groupby(RecLevelID)])
        L.append(df)
        
    # Merge the dataframes in the list using reduce
    df_data = reduce(lambda left, right: pd.merge(left, right, on=RecLevelID, how = 'left'), L)

    # 3. get df_whole
    df_whole = pd.merge(df_prefix, df_data, on = RecLevelID, how = 'left')

    for RecName in RecTableName2FldColumns_Dict:
        ind = df_whole[RecName].isna()
        columns = df_whole[-ind][RecName].iloc[0].columns
        df_whole[RecName] = df_whole[RecName].apply(lambda x: x if type(x) == type(pd.DataFrame()) 
                                                    else pd.DataFrame(columns = columns))
    
    return df_whole


def get_df_bucket_from_settings(bucket_file, RecChain_ARGS, RecInfo_ARGS):
    # 1 ----------------- get df_prefix
    L = []
    # RecNameID_Chain = [f'{i}ID' for i in RecName_Chain]
    RecNameID_Chain = [i for i in RecChain_ARGS]
    for idx, RID in enumerate(RecChain_ARGS):
        RecInfo = RecChain_ARGS[RID]
        # folder, Name = RecName_Chain_To_RecFolderName.get(RecName, (rec_folder, RecName))
        folder, RecName = RecInfo['folder'], RecInfo['RecName']
        df = pd.read_pickle(os.path.join(folder, RecName, bucket_file))
        df = df[RecNameID_Chain[:idx+ 1]].astype(str).drop_duplicates()
        L.append(df)
        
    df_prefix = reduce(lambda left, right: pd.merge(left, right, how = 'left'), L)
    # print(df_prefix)
    
    # fill the missing Rec with missing ID.
    for RID in RecNameID_Chain:
        s = 'M' + pd.Series(df_prefix.index).astype(str)
        df_prefix[RID] = df_prefix[RID].fillna(s)

    
    # 2 ----------------- get df_data
    # RecLevel = RecNameID_Chain[-1][:-2]
    # RecLevelID = f'{RecLevel}ID'
    RecLevelID = RecNameID_Chain[-1]

    L = []
    # for RecName, FldList in RecTableName2FldColumns_Dict.items():
    for idx, RecElmt in enumerate(RecInfo_ARGS):
        RecElmt_ARGS = RecInfo_ARGS[RecElmt]
        folder = RecElmt_ARGS['folder']
        RecName = RecElmt_ARGS['RecName']
        FldList = RecElmt_ARGS['Columns']
        
        # read df
        path = os.path.join(folder, RecName, bucket_file)
        if not os.path.exists(path): print(f'empty path: {path}'); continue
        df = pd.read_pickle(path)
        
        # select columns
        if FldList == 'ALL': FldList = list(df.columns)
        full_cols = [i for i in RecNameID_Chain if i not in FldList] + FldList
        full_cols = [i for i in full_cols if i in df.columns]
        df = df[full_cols].reset_index(drop = True)
        
        # update columns
        for RecID in RecNameID_Chain:
            if RecID in df.columns: df[RecID] = df[RecID].astype(str)

        # downstream df_data with df_prefix if RecLevelID is not in ID
        # eg: RecLevelID is Encounter-level, but df is Patient-level.
        if RecLevelID not in df.columns:
            on_cols = [i for i in df_prefix.columns if i in df.columns]
            df = pd.merge(df_prefix, df, on = on_cols, how = 'left')

        # change df to RecLevelID for each row.
        df = pd.DataFrame([{RecLevelID: RecLevelIDValue, RecElmt: df_input} 
                           for RecLevelIDValue, df_input in df.groupby(RecLevelID)])
        
        L.append(df)
        
    # uncomment this is you want to check
    # for df in L: print(df.shape)
    
    # 3 ----------------- get df_data
    # Merge the dataframes in the list using reduce, here we should use OUTER!
    # <!!!!!!!!!!!!!!!! here is the data only, so we use outer !!!!!!!!!!!!!!!!>
    df_data = reduce(lambda left, right: pd.merge(left, right, on=RecLevelID, how = 'outer'), L)

    # 4 ----------------- get df_whole
    # <!!!!!!!!!!!!!!!! here we only focus rows in df_prefix, so we use left !!!!!!!!!!!!!!!!>
    df_whole = pd.merge(df_prefix, df_data, on = RecLevelID, how = 'left')
    
    return df_whole


import os 
import random
import pandas as pd

RANGE_SIZE = 10000 # This need to be adjusted for different RecFld Folder.

# Convert PID to PID Group
def convert_PID_to_PIDgroup(x, RANGE_SIZE):
    # x here is the PID, it can be a string (P0002) or int.
    if type(x) == str:
        if 'P' in x:
            int_value = int(x.replace('P', ''))
        else:
            int_value = int(x)
    else:
        int_value = int(x)
    range_size = RANGE_SIZE
    group_idx = int(int_value / range_size)
    s = group_idx * range_size
    e = (group_idx + 1) * range_size - 1
    groupname = f'Group_s{s}_e{e}'
    return groupname

# write df to folder
def write_df_to_folders(data_folder, df, IDNAME, RANGE_SIZE = RANGE_SIZE):
    dfy = df.copy()
    if not os.path.exists(data_folder):
        os.makedirs(data_folder)

    dfy['GroupName'] = dfy[IDNAME].apply(lambda x: convert_PID_to_PIDgroup(x , RANGE_SIZE))
    for groupname, dfx in dfy.groupby('GroupName'):
        file = os.path.join(data_folder, f'{groupname}.p')
        # print(file)
        dfx = dfx.drop(columns= ['GroupName'])
        dfx.to_pickle(file)
        
# load df from folder
def load_df_data_from_folder(data_folder, **kwargs):
    # data_folder: = recfld_folder + RecName
    # rec_folder: the folder to hold many Rec types.
    file_list = os.listdir(data_folder)
    df = pd.concat([pd.read_pickle(os.path.join(data_folder, i)) for i in file_list if '.p' in i]).reset_index(drop = True)
    return df


def write_df_data_with_RecName(df, rec_folder, RecName, IDNAME, RANGE_SIZE = RANGE_SIZE):
    fullrec_folder = os.path.join(rec_folder, RecName)
    if not os.path.exists(fullrec_folder): os.makedirs(fullrec_folder)

    dfy = df.copy()
    dfy['GroupName'] = dfy[IDNAME].apply(lambda x: convert_PID_to_PIDgroup(x , RANGE_SIZE))
    for groupname, dfx in dfy.groupby('GroupName'):
        file = os.path.join(fullrec_folder, f'{groupname}.p')
        dfx = dfx.drop(columns= ['GroupName'])
        dfx.to_pickle(file)

def load_df_data_with_RecName(rec_folder, RecName):
    data_folder = os.path.join(rec_folder, RecName)
    file_list = sorted(os.listdir(data_folder))
    df = pd.concat([pd.read_pickle(os.path.join(data_folder, i)) for i in file_list if '.p' in i]).reset_index(drop = True)
    return df

def get_dfx_p_filename(IDValue, RecName, rec_folder, RANGE_SIZE):
    group_name = convert_PID_to_PIDgroup(IDValue, RANGE_SIZE) 
    p_filename = f'{rec_folder}/{RecName}/{group_name}.p'
    return p_filename
    
def load_dp_data(IDName, IDValue, dfx):
    # current Jul 15, 2023: All IDName must be 'PID', all IDValue is PID's value
    # IDName: PID, IDValue: value of PID
    dfx = dfx[dfx[IDName] == IDValue].reset_index(drop = True)
    return dfx

def get_dfx_from_buffer(IDName, IDValue, DataName, data_folder, RANGE_SIZE, BUCKET_buffer):
    p_filename = get_dfx_p_filename(IDValue, DataName, data_folder, RANGE_SIZE)
    if p_filename not in BUCKET_buffer:
        dfx = pd.read_pickle(p_filename)
        BUCKET_buffer[p_filename] = dfx
        if len(BUCKET_buffer) > 100:
            for key in random.sample(BUCKET_buffer.keys(), 4):
                del BUCKET_buffer[key]
    else:
        dfx = BUCKET_buffer[p_filename]
    return dfx, BUCKET_buffer


class DataPoint(object):

    '''
    DataPoint: stands for one data point. 
    one instance is a patient.
    the instance of the patient could quickly the his/her record information by using `get_df_rec`.
    '''
    BUCKET_buffer = {}

    def __init__(self, IDName, IDValue, rec_folder, RANGE_SIZE):
        self.IDName = IDName
        self.IDValue = IDValue
        self.rec_folder = rec_folder
        self.df_rec_dict = {}
        self.RANGE_SIZE = RANGE_SIZE
        
    def get_df_rec(self, RecName):
        '''
        RecName: the name of the record, like: P, EC, BMI, PN, PNSect, etc.
        '''
        if RecName not in self.df_rec_dict:
            dfx, BUCKET_buffer = get_dfx_from_buffer(self.IDName, self.IDValue, 
                                                     RecName, 
                                                     self.rec_folder, 
                                                     self.RANGE_SIZE, 
                                                     self.BUCKET_buffer)
            self.BUCKET_buffer = BUCKET_buffer
            df_p = load_dp_data(self.IDName, self.IDValue, dfx)
            self.df_rec_dict[RecName] = df_p
            return df_p
        else:
            return self.df_rec_dict[RecName]
        
    def __repr__(self):
        return f'<{self.IDName} [{self.IDValue}]>'
    


from setuptools import setup, find_packages

setup(
    name='recfldgrn',
    version='0.0.1',
    description='Your package description',
    author='Your Name',
    author_email='junjie.luo.jhu@gmail.com',
    packages=find_packages(),  # Automatically find all packages
)


def assign_dfwhole_obs_pred_fut_date(df_whole, DTtable, DTcol, BfRange = None, AfRange = None):
    predidx_date = df_whole[DTtable].apply(lambda x: x[DTcol].iloc[0])
    df_whole['predind_date'] = predidx_date
    
    RangeList = [i for i in [BfRange, AfRange] if i != None]
    
    if len(RangeList) == 0: 
        raise ValueError(f'Range Options is zero, you need to provide at least one of Bf and Af')
       
    for Range in RangeList:
        RangeWhSign = Range.replace('BF', '-').replace('AF', '')
        if RangeWhSign[-1] == 'y':
            number = int(RangeWhSign.replace('y', ''))
            days = number * 365
        elif RangeWhSign[-1] == 'd':
            number = int(RangeWhSign.replace('d', ''))
            days = number
        else:
            raise ValueError(f'The unit in Range {Range} is not available yet')

        bound_date = predidx_date.apply(lambda x: x + pd.to_timedelta(days, unit = 'days'))
        if number > 0: 
            df_whole['futind_date'] = bound_date; 
        else:  
            df_whole['obsind_date'] = bound_date; 

    return df_whole


def filter_dfwhole_content_with_start_end_date(df_whole, start_col, end_col, Rec2TimeCols_dfwhole_content):
    
    for Rec, TimeCols in Rec2TimeCols_dfwhole_content.items():
        for TimeCol in TimeCols:
            df_whole['ts'] = df_whole[Rec].apply(lambda x: x[TimeCol].iloc[0])
            # TODO: how to customized this index generation.
            index = - (df_whole.apply(lambda row: row['ts'] < row[start_col] or row['ts'] > row[end_col], axis = 1))
            df_whole = df_whole[index].reset_index(drop = True)
             
    df_whole = df_whole.drop(columns = ['ts'])
    
    return df_whole


def filter_row_content_with_start_end_date(row, start_col, end_col, Rec2TimeCols_row_content):
    
    for Rec, TimeCols_list in Rec2TimeCols_row_content.items():
        
        for TimeCol in TimeCols_list:
            df = row[Rec]; df['ts'] = df[TimeCol]
            # TODO: how to customized this index generation. 
            index = - (df.apply(lambda x: x['ts'] < row[start_col] or x['ts'] > row[end_col], axis = 1)) # considering missing records
            df = df[index].reset_index(drop = True)
            df = df.drop(columns = ['ts']); row[Rec] = df
        
    return row



import os
import numpy as np
import pandas as pd
from functools import reduce
from .datapoint import load_df_data_from_folder
import itertools


def func_convert_Num_factory(start, end, Min, Max, scale):
    
    def func_convert_NUMgrn(v):

        miss, more, less, bottom, top = 0, 0, 0, 0, 0; base = 1
        
        if pd.isna(v) == True:
            miss = 1; base = 0
            num_embed = int((end - start) / scale)
            x = np.zeros(num_embed + 1)# .astype(int)
            x = list(x)
            wgt = [miss, more, less, bottom, top, base] + x
            tkn = ['miss', 'more', 'less', 'bottom', 'top', 'base'] + [f'B{start + i*scale}+' for i in range(len(x))]
            return tkn, wgt
        
        else:
            if type(v) == str:
                if v[0] not in '<>-': v = float(v) # No special symbols
                elif v[0] == '<': less = 1; v = float(v[1:]) # if < is here
                elif v[0] == '>': more = 1; v = float(v[1:]) # if > is here
                
            if v < start:  bottom = (start - v) / (start - Min); base = 0; v = start
            elif v > end: top = (v-end) / (Max-end); v = end

            num_embed = int((end - start) / scale)
            x = np.zeros(num_embed + 1)# .astype(f)
            num = int((v - start) / scale)
            # print(num)
            x[:num] = 1
            # print(v, start, scale,  v - start, num)
            # print( (v-start) - num * scale)
            x[num] = ((v-start) - num * scale) / scale
            x = x.round(4)
            # print((v - start) % scale, v, start, scale)
            # print(x[num])
            # print(x)
            x = list(x) # value part

            wgt = [miss, more, less, bottom, top, base] + x
            tkn = ['miss', 'more', 'less', 'bottom', 'top', 'base'] +[f'B{start + i*scale}+' for i in range(len(x))]
            return tkn, wgt
    return func_convert_NUMgrn

def generate_grain_vocab_info(s):
    '''
        s: a Series (or a column of a dataframe) of all rec@fld value's grain_list.
            eg. grain_list: a list of tokens from raw sentence. 
        
        output: a dict of idx2v, v2dix, v2freq.
    '''
    value_list = itertools.chain(*s.to_list())
    d = pd.Series(value_list).value_counts().to_dict()
    v2freq = {'_padding': 0}
    if '_missing' in d:
        v2freq['_missing'] = d.pop('_missing') # I should have some prefix here. 
    for k, v in d.items(): v2freq[k] = v

    idx2v = {idx: v for idx, v in enumerate(v2freq.keys())}
    v2idx = {v:k for k, v in idx2v.items()}
    
    return {'idx2v': idx2v, 'v2idx': v2idx, 'v2freq': v2freq}


def get_compressed_df(df_rec, full_recfldgrn_name, prefix_ids):
    '''
        df_rec: with columns: prefix_ids + focal_ids + full_recfldgrn_name
            prefix_ids: [PID, ECID] 
            focal_ids: [PNID], current primary key
            full_recfldgrn_name: `PrimaryNote - Section - SectSent@Sentence - TkGrn`

            prefix_ids_new: [PID] 
            focal_ids_new: [ECID], current primary key
            full_recfldgrn_name_new: `EC - PrimaryNote - Section - SectSent@Sentence - TkGrn`
    '''
    CompressParent_ID_list = prefix_ids # parent layer's ID which will be used to compress. 
    CompressParent_ID = prefix_ids[-1]
    prefix_ids_new = prefix_ids[:-1]   # the new prefix_ids. 

    df_rec_new = pd.DataFrame(df_rec.groupby(CompressParent_ID_list).apply(lambda x: x.to_dict('list')).to_list())

    for col in prefix_ids: 
        df_rec_new[col] = df_rec_new[col].apply(lambda x: x[0])
        
    full_recfldgrn_name_new = CompressParent_ID.replace('ID', '') + '-' + full_recfldgrn_name
    df_rec_new = df_rec_new.rename(columns = {full_recfldgrn_name: full_recfldgrn_name_new})
    df_rec_new = df_rec_new[prefix_ids + [full_recfldgrn_name_new]]
    return df_rec_new, full_recfldgrn_name_new, prefix_ids_new


def get_highorder_input_idx(df, recfldgrn_sfx, prefix_ids, focal_ids):
    '''
        df: raw data with column: prefix_ids + focal_ids + [recfldgrn_sfx]
        recfldgrn_sfx: xxx_wgt, or xxx_tknidx, or xxx_fldidx
        prefix_ids: [PID, ECID, PNID, PNSectID]
        focal_ids: [SectSentID]
    '''
    # recfldgrn = recfldgrn_sfx.split('_')[0]
    # rec, field = recfldgrn.split('-')[0].split('@')
    # grain = recfldgrn.split('-')[-1]

    # top_id = df.columns[0] # need to double check this.
    if len(prefix_ids) > 0:
        prefix_ids_c = prefix_ids.copy() # _c: copy (meaningless)
        df_Rec = df[prefix_ids_c + [recfldgrn_sfx]]
        for i in range(len(prefix_ids_c)):
            df_Rec, recfldgrn_sfx, prefix_ids_c = get_compressed_df(df_Rec, recfldgrn_sfx, prefix_ids_c)
    else:
        df_Rec = df[focal_ids + [recfldgrn_sfx]]

    df_p = df_Rec.reset_index(drop = True) # p: patient.
    return df_p


def generate_EmbedDict_from_processed_df_whole(df_whole, recfldgrn, sfx_list = ['tkn', 'tpc']):
    EmbedDict = {}
    
    for sfx in sfx_list:
        embed_type = 'CateEmbed'
        recfldgrn_sfx = f'{recfldgrn}_{sfx}'
        # print(recfldgrn_sfx)
        s = df_whole[recfldgrn_sfx] 
        Vocab = generate_grain_vocab_info(s) 
        v2idx = Vocab['v2idx']
        vocab_size = len(v2idx)
        # print(v2idx)
        d = {'embed_type':embed_type, 
             'recfldgrn_sfx':recfldgrn_sfx,
             'vocab_size': vocab_size,
             'Vocab': Vocab}
        
        df_whole[f'{recfldgrn}_{sfx}idx'] =  df_whole[f'{recfldgrn}_{sfx}'].apply(lambda x: [v2idx[i] for i in x])
        EmbedDict[sfx] = d
        
    return df_whole, EmbedDict
